@@ -10,18 +10,17 @@
 -- files. Run `luarocks` with no arguments to see the locations of
 -- these files in your platform.
 
-local next, table, pairs, require, os, pcall, ipairs, package, tonumber, type, assert =
-      next, table, pairs, require, os, pcall, ipairs, package, tonumber, type, assert
+local table, pairs, require, os, pcall, ipairs, package, type, assert =
+      table, pairs, require, os, pcall, ipairs, package, type, assert
 
 local util = require("luarocks.core.util")
 local persist = require("luarocks.core.persist")
 local sysdetect = require("luarocks.core.sysdetect")
+local vers = require("luarocks.core.vers")
 
 --------------------------------------------------------------------------------
 
-local program_version = "3.1.3"
-local program_series = "3.1"
-local major_version = (program_version:match("([^.]%.[^.])")) or program_series
+local program_version = "3.9.2"
 
 local is_windows = package.config:sub(1,1) == "\\"
 
@@ -36,6 +35,7 @@ local platform_order = {
    "netbsd",
    "openbsd",
    "freebsd",
+   "dragonfly",
    "linux",
    "macosx",
    "cygwin",
@@ -44,10 +44,15 @@ local platform_order = {
    -- Windows
    "windows",
    "win32",
+   "mingw",
    "mingw32",
+   "msys2_mingw_w64",
 }
 
 local function detect_sysconfdir()
+   if not debug then
+      return
+   end
    local src = debug.getinfo(1, "S").source:gsub("\\", "/"):gsub("/+", "/")
    if src:sub(1, 1) == "@" then
       src = src:sub(2)
@@ -68,35 +73,20 @@ local function detect_sysconfdir()
    return basedir
 end
 
-local function set_confdirs(cfg, platforms, hardcoded)
-   local sysconfdir = os.getenv("LUAROCKS_SYSCONFDIR") or hardcoded.SYSCONFDIR
-   if platforms.windows then
-      cfg.home = os.getenv("APPDATA") or "c:"
-      cfg.home_tree = cfg.home.."/luarocks"
-      cfg.homeconfdir = cfg.home_tree
-      cfg.sysconfdir = sysconfdir or ((os.getenv("PROGRAMFILES") or "c:") .. "/luarocks")
-   else
-      if not sysconfdir then
-         sysconfdir = detect_sysconfdir()
-      end
-      cfg.home = hardcoded.HOMEDIR or os.getenv("HOME") or ""
-      cfg.localdir = hardcoded.LOCALDIR or cfg.home
-      local home_tree_subdir = hardcoded.HOME_TREE_SUBDIR or "/.luarocks"
-      cfg.homeconfdir = cfg.localdir .. home_tree_subdir
-      cfg.sysconfdir = sysconfdir or "/etc/luarocks"
-      cfg.home_tree = cfg.localdir .. home_tree_subdir
-   end
-end
-
 local load_config_file
 do
    -- Create global environment for the config files;
    local function env_for_config_file(cfg, platforms)
+      local platforms_copy = {}
+      for k,v in pairs(platforms) do
+         platforms_copy[k] = v
+      end
+
       local e
       e = {
          home = cfg.home,
          lua_version = cfg.lua_version,
-         platforms = util.make_shallow_copy(platforms),
+         platforms = platforms_copy,
          processor = cfg.target_cpu,   -- remains for compat reasons
          target_cpu = cfg.target_cpu,  -- replaces `processor`
          os_getenv = os.getenv,
@@ -160,6 +150,7 @@ end
 local platform_sets = {
    freebsd = { unix = true, bsd = true, freebsd = true },
    openbsd = { unix = true, bsd = true, openbsd = true },
+   dragonfly = { unix = true, bsd = true, dragonfly = true },
    solaris = { unix = true, solaris = true },
    windows = { windows = true, win32 = true },
    cygwin = { unix = true, cygwin = true },
@@ -169,6 +160,7 @@ local platform_sets = {
    linux = { unix = true, linux = true },
    mingw = { windows = true, win32 = true, mingw32 = true, mingw = true },
    msys = { unix = true, cygwin = true, msys = true },
+   msys2_mingw_w64 = { windows = true, win32 = true, mingw32 = true, mingw = true, msys = true, msys2_mingw_w64 = true },
 }
 
 local function make_platforms(system)
@@ -189,11 +181,11 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
       fs_use_modules = true,
       hooks_enabled = true,
       deps_mode = "one",
+      no_manifest = false,
       check_certificates = false,
 
       cache_timeout = 60,
       cache_fail_timeout = 86400,
-      version_check_on_fail = true,
 
       lua_modules_path = hardcoded.LUA_MODULES_LUA_SUBDIR or "/share/lua/"..lua_version,
       lib_modules_path = hardcoded.LUA_MODULES_LIB_SUBDIR or "/lib/lua/"..lua_version,
@@ -208,8 +200,7 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
          {
            "https://luarocks.org",
            "https://raw.githubusercontent.com/rocks-moonscript-org/moonrocks-mirror/master/",
-           "http://luafr.org/moonrocks/",
-           "http://luarocks.logiceditor.com/rocks",
+           "https://luafr.org/luarocks/",
          }
       },
       disabled_servers = {},
@@ -224,11 +215,11 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
       connection_timeout = 30,  -- 0 = no timeout
 
       variables = {
-         MAKE = "make",
-         CC = "cc",
-         LD = "ld",
-         AR = "ar",
-         RANLIB = "ranlib",
+         MAKE = os.getenv("MAKE") or "make",
+         CC = os.getenv("CC") or "cc",
+         LD = os.getenv("CC") or "ld",
+         AR = os.getenv("AR") or "ar",
+         RANLIB = os.getenv("RANLIB") or "ranlib",
 
          CVS = "cvs",
          GIT = "git",
@@ -295,14 +286,16 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
       defaults.external_deps_dirs = { "c:/external/", "c:/windows/system32" }
 
       defaults.makefile = "Makefile.win"
-      defaults.variables.MAKE = "nmake"
-      defaults.variables.CC = "cl"
-      defaults.variables.RC = "rc"
-      defaults.variables.LD = "link"
-      defaults.variables.MT = "mt"
-      defaults.variables.AR = "lib"
+      defaults.variables.PWD = "echo %cd%"
+      defaults.variables.MAKE = os.getenv("MAKE") or "nmake"
+      defaults.variables.CC = os.getenv("CC") or "cl"
+      defaults.variables.RC = os.getenv("WINDRES") or "rc"
+      defaults.variables.LD = os.getenv("LINK") or "link"
+      defaults.variables.MT = os.getenv("MT") or "mt"
+      defaults.variables.AR = os.getenv("AR") or "lib"
       defaults.variables.LUALIB = "lua"..lua_version..".lib"
-      defaults.variables.CFLAGS = "/nologo /MD /O2"
+      defaults.variables.CFLAGS = os.getenv("CFLAGS") or "/nologo /MD /O2"
+      defaults.variables.LDFLAGS = os.getenv("LDFLAGS")
       defaults.variables.LIBFLAG = "/nologo /dll"
 
       defaults.external_deps_patterns = {
@@ -337,13 +330,19 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
       defaults.static_lib_extension = "a"
       defaults.external_deps_dirs = { "c:/external/", "c:/mingw", "c:/windows/system32" }
       defaults.cmake_generator = "MinGW Makefiles"
-      defaults.variables.MAKE = "mingw32-make"
-      defaults.variables.CC = "mingw32-gcc"
-      defaults.variables.RC = "windres"
-      defaults.variables.LD = "mingw32-gcc"
-      defaults.variables.AR = "ar"
-      defaults.variables.RANLIB = "ranlib"
-      defaults.variables.CFLAGS = "-O2"
+      defaults.variables.MAKE = os.getenv("MAKE") or "mingw32-make"
+      if target_cpu == "x86_64" then
+         defaults.variables.CC = os.getenv("CC") or "x86_64-w64-mingw32-gcc"
+         defaults.variables.LD = os.getenv("CC") or "x86_64-w64-mingw32-gcc"
+      else
+         defaults.variables.CC = os.getenv("CC") or "mingw32-gcc"
+         defaults.variables.LD = os.getenv("CC") or "mingw32-gcc"
+      end
+      defaults.variables.AR = os.getenv("AR") or "ar"
+      defaults.variables.RC = os.getenv("WINDRES") or "windres"
+      defaults.variables.RANLIB = os.getenv("RANLIB") or "ranlib"
+      defaults.variables.CFLAGS = os.getenv("CFLAGS") or "-O2"
+      defaults.variables.LDFLAGS = os.getenv("LDFLAGS")
       defaults.variables.LIBFLAG = "-shared"
       defaults.makefile = "Makefile"
       defaults.external_deps_patterns = {
@@ -358,6 +357,7 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
          lib = { "cyg?.dll", "?.dll", "lib?.dll" },
          include = { "?.h" }
       }
+      defaults.link_lua_explicitly = true
    end
 
    if platforms.unix then
@@ -366,12 +366,22 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
       defaults.external_lib_extension = "so"
       defaults.obj_extension = "o"
       defaults.external_deps_dirs = { "/usr/local", "/usr", "/", hardcoded.PREFIX }
-      defaults.variables.CFLAGS = "-O2"
+
+      defaults.variables.CFLAGS = os.getenv("CFLAGS") or "-O2"
+      -- we pass -fPIC via CFLAGS because of old Makefile-based Lua projects
+      -- which didn't have -fPIC in their Makefiles but which honor CFLAGS
+      if not defaults.variables.CFLAGS:match("-fPIC") then
+         defaults.variables.CFLAGS = defaults.variables.CFLAGS.." -fPIC"
+      end
+
+      defaults.variables.LDFLAGS = os.getenv("LDFLAGS")
+
       defaults.cmake_generator = "Unix Makefiles"
-      defaults.variables.CC = "gcc"
-      defaults.variables.LD = "gcc"
+      defaults.variables.CC = os.getenv("CC") or "gcc"
+      defaults.variables.LD = os.getenv("CC") or "gcc"
       defaults.gcc_rpath = true
       defaults.variables.LIBFLAG = "-shared"
+      defaults.variables.TEST = "test"
 
       defaults.external_deps_patterns = {
          bin = { "?" },
@@ -385,10 +395,8 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
       }
       defaults.export_path_separator = ":"
       defaults.wrapper_suffix = ""
-      defaults.local_cache = home.."/.cache/luarocks"
-      if not defaults.variables.CFLAGS:match("-fPIC") then
-         defaults.variables.CFLAGS = defaults.variables.CFLAGS.." -fPIC"
-      end
+      local xdg_cache_home = os.getenv("XDG_CACHE_HOME") or home.."/.cache"
+      defaults.local_cache = xdg_cache_home.."/luarocks"
       defaults.web_browser = "xdg-open"
    end
 
@@ -396,8 +404,8 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
       defaults.lib_extension = "so" -- can be overridden in the config file for mingw builds
       defaults.arch = "cygwin-"..target_cpu
       defaults.cmake_generator = "Unix Makefiles"
-      defaults.variables.CC = "echo -llua | xargs gcc"
-      defaults.variables.LD = "echo -llua | xargs gcc"
+      defaults.variables.CC = "echo -llua | xargs " .. (os.getenv("CC") or "gcc")
+      defaults.variables.LD = "echo -llua | xargs " .. (os.getenv("CC") or "gcc")
       defaults.variables.LIBFLAG = "-shared"
       defaults.link_lua_explicitly = true
    end
@@ -416,29 +424,80 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
          lib = { "lib?.so", "?.dll", "lib?.dll" },
          include = { "?.h" }
       }
+      if platforms.mingw then
+         -- MSYS2 can build Windows programs that depend on
+         -- msys-2.0.dll (based on Cygwin) but MSYS2 is also designed
+         -- for building native Windows programs by MinGW. These
+         -- programs don't depend on msys-2.0.dll.
+         local pipe = io.popen("cygpath --windows %MINGW_PREFIX%")
+         local mingw_prefix = pipe:read("*l")
+         pipe:close()
+         defaults.external_deps_dirs = { mingw_prefix, "c:/windows/system32" }
+         defaults.makefile = "Makefile"
+         defaults.cmake_generator = "MSYS Makefiles"
+         defaults.local_cache = home.."/.cache/luarocks"
+         defaults.variables.MAKE = os.getenv("MAKE") or "make"
+         defaults.variables.CC = os.getenv("CC") or "gcc"
+         defaults.variables.RC = os.getenv("WINDRES") or "windres"
+         defaults.variables.LD = os.getenv("CC") or "gcc"
+         defaults.variables.MT = os.getenv("MT") or nil
+         defaults.variables.AR = os.getenv("AR") or "ar"
+         defaults.variables.RANLIB = os.getenv("RANLIB") or "ranlib"
+         defaults.variables.LUALIB = "liblua"..lua_version..".dll.a"
+
+         defaults.variables.CFLAGS = os.getenv("CFLAGS") or "-O2 -fPIC"
+         if not defaults.variables.CFLAGS:match("-fPIC") then
+            defaults.variables.CFLAGS = defaults.variables.CFLAGS.." -fPIC"
+         end
+
+         defaults.variables.LIBFLAG = "-shared"
+      end
    end
 
    if platforms.bsd then
       defaults.variables.MAKE = "gmake"
+      defaults.gcc_rpath = false
+      defaults.variables.CC = os.getenv("CC") or "cc"
+      defaults.variables.LD = os.getenv("CC") or defaults.variables.CC
    end
 
    if platforms.macosx then
-      defaults.variables.MAKE = "make"
+      defaults.variables.MAKE = os.getenv("MAKE") or "make"
       defaults.external_lib_extension = "dylib"
       defaults.arch = "macosx-"..target_cpu
       defaults.variables.LIBFLAG = "-bundle -undefined dynamic_lookup -all_load"
       local version = util.popen_read("sw_vers -productVersion")
-      version = tonumber(version and version:match("^[^.]+%.([^.]+)")) or 3
-      if version >= 10 then
-         version = 8
-      elseif version >= 5 then
-         version = 5
+      if not (version:match("^%d+%.%d+%.%d+$") or version:match("^%d+%.%d+$")) then
+         version = "10.3"
+      end
+      version = vers.parse_version(version)
+      if version >= vers.parse_version("11.0") then
+         version = vers.parse_version("11.0")
+      elseif version >= vers.parse_version("10.10") then
+         version = vers.parse_version("10.8")
+      elseif version >= vers.parse_version("10.5") then
+         version = vers.parse_version("10.5")
       else
          defaults.gcc_rpath = false
       end
-      defaults.variables.CC = "env MACOSX_DEPLOYMENT_TARGET=10."..version.." gcc"
-      defaults.variables.LD = "env MACOSX_DEPLOYMENT_TARGET=10."..version.." gcc"
+      defaults.variables.CC = "env MACOSX_DEPLOYMENT_TARGET="..tostring(version).." gcc"
+      defaults.variables.LD = "env MACOSX_DEPLOYMENT_TARGET="..tostring(version).." gcc"
       defaults.web_browser = "open"
+
+      -- XCode SDK
+      local sdk_path = util.popen_read("xcrun --show-sdk-path 2>/dev/null")
+      if sdk_path then
+         table.insert(defaults.external_deps_dirs, sdk_path .. "/usr")
+         table.insert(defaults.external_deps_patterns.lib, 1, "lib?.tbd")
+         table.insert(defaults.runtime_external_deps_patterns.lib, 1, "lib?.tbd")
+      end
+
+      -- Homebrew
+      table.insert(defaults.external_deps_dirs, "/usr/local/opt")
+      defaults.external_deps_subdirs.lib = { "", "lib", }
+      defaults.runtime_external_deps_subdirs.lib = { "", "lib", }
+      table.insert(defaults.external_deps_patterns.lib, 1, "/?/lib/lib?.dylib")
+      table.insert(defaults.runtime_external_deps_patterns.lib, 1, "/?/lib/lib?.dylib")
    end
 
    if platforms.linux then
@@ -456,22 +515,14 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
 
    if platforms.freebsd then
       defaults.arch = "freebsd-"..target_cpu
-      defaults.gcc_rpath = false
-      defaults.variables.CC = "cc"
-      defaults.variables.LD = "cc"
-   end
-
-   if platforms.openbsd then
+   elseif platforms.dragonfly then
+      defaults.arch = "dragonfly-"..target_cpu
+   elseif platforms.openbsd then
       defaults.arch = "openbsd-"..target_cpu
-   end
-
-   if platforms.netbsd then
+   elseif platforms.netbsd then
       defaults.arch = "netbsd-"..target_cpu
-   end
-
-   if platforms.solaris then
+   elseif platforms.solaris then
       defaults.arch = "solaris-"..target_cpu
-      --defaults.platforms = {"unix", "solaris"}
       defaults.variables.MAKE = "gmake"
    end
 
@@ -482,50 +533,18 @@ local function make_defaults(lua_version, target_cpu, platforms, home, hardcoded
    return defaults
 end
 
-local function make_rocks_provided(lua_version, luajit_version)
-   local rocks_provided = {}
-   local rocks_provided_3_0 = {}
-
-   rocks_provided["lua"] = lua_version.."-1"
-
-   if lua_version == "5.2" or lua_version == "5.3" then
-      rocks_provided["bit32"] = lua_version.."-1"
-   end
-
-   if lua_version == "5.3" or lua_version == "5.4" then
-      rocks_provided["utf8"] = lua_version.."-1"
-   end
-
-   if luajit_version then
-      rocks_provided["luabitop"] = luajit_version.."-1"
-      rocks_provided_3_0["luajit"] = luajit_version.."-1"
-   end
-
-   if rawget(_G, '_TARANTOOL') then
-      -- Tarantool
-      local tarantool_version = _TARANTOOL:match("([^-]+)-")
-      rocks_provided["tarantool"] = tarantool_version.."-1"
-      rocks_provided_3_0["tarantool"] = tarantool_version.."-1"
-   end
-
-   return rocks_provided, rocks_provided_3_0
-end
-
 local function use_defaults(cfg, defaults)
 
-   -- Populate some arrays with values from their 'defaults' counterparts
+   -- Populate variables with values from their 'defaults' counterparts
    -- if they were not already set by user.
-   for _, entry in ipairs({"variables", "rocks_provided"}) do
-      if not cfg[entry] then
-         cfg[entry] = {}
-      end
-      for k,v in pairs(defaults[entry]) do
-         if not cfg[entry][k] then
-            cfg[entry][k] = v
-         end
+   if not cfg.variables then
+      cfg.variables = {}
+   end
+   for k,v in pairs(defaults.variables) do
+      if not cfg.variables[k] then
+         cfg.variables[k] = v
       end
    end
-   util.deep_merge_under(defaults.rocks_provided_3_0, cfg.rocks_provided)
 
    util.deep_merge_under(cfg, defaults)
 
@@ -536,19 +555,30 @@ local function use_defaults(cfg, defaults)
    end
 end
 
+local function get_first_arg()
+   local arg = rawget(_G, 'arg')
+   if not arg then
+      return
+   end
+   local first_arg = arg[0]
+   local i = -1
+   while arg[i] do
+      first_arg = arg[i]
+      i = i -1
+   end
+   return first_arg
+end
+
 --------------------------------------------------------------------------------
 
 local cfg = {}
 
 --- Initializes the LuaRocks configuration for variables, paths
 -- and OS detection.
--- @param detected table containing information detected about the 
+-- @param detected table containing information detected about the
 -- environment. All fields below are optional:
 -- * lua_version (in x.y format, e.g. "5.3")
--- * luajit_version (complete, e.g. "2.1.0-beta3")
 -- * lua_bindir (e.g. "/usr/local/bin")
--- * lua_incdir (e.g. "/usr/local/include/lua5.3/")
--- * lua_libdir(e.g. "/usr/local/lib")
 -- * lua_dir (e.g. "/usr/local")
 -- * lua_interpreter (e.g. "lua-5.3")
 -- * project_dir (a string with the path of the project directory
@@ -562,21 +592,15 @@ function cfg.init(detected, warning)
 
    detected = detected or {}
 
+   local exit_ok = true
+   local exit_err = nil
+   local exit_what = nil
+
    local hc_ok, hardcoded = pcall(require, "luarocks.core.hardcoded")
    if not hc_ok then
       hardcoded = {}
    end
 
-   local arg = rawget(_G, 'arg')
-   local lua_version = detected.lua_version or hardcoded.LUA_VERSION or _VERSION:sub(5)
-   local luajit_version = detected.luajit_version or hardcoded.LUAJIT_VERSION or (jit and jit.version:sub(8))
-   local lua_interpreter = detected.lua_interpreter or hardcoded.LUA_INTERPRETER or (arg and arg[-1] and arg[-1]:gsub(".*[\\/]", "")) or (is_windows and "lua.exe" or "lua")
-   local lua_bindir = detected.lua_bindir or hardcoded.LUA_BINDIR or (arg and arg[-1] and arg[-1]:gsub("[\\/][^\\/]+$", ""))
-   local lua_incdir = detected.lua_incdir or hardcoded.LUA_INCDIR
-   local lua_libdir = detected.lua_libdir or hardcoded.LUA_LIBDIR
-   local lua_dir = detected.lua_dir or hardcoded.LUA_DIR or (lua_bindir and lua_bindir:gsub("[\\/]bin$", ""))
-   local project_dir = detected.project_dir
-   
    local init = cfg.init
 
    ----------------------------------------
@@ -588,19 +612,29 @@ function cfg.init(detected, warning)
    end
 
    cfg.program_version = program_version
-   cfg.program_series = program_series
-   cfg.major_version = major_version
 
-   cfg.lua_version = lua_version
-   cfg.luajit_version = luajit_version
-   cfg.lua_interpreter = lua_interpreter
+   if hardcoded.IS_BINARY then
+      cfg.is_binary = true
+   end
 
-   cfg.variables = {
-      LUA_DIR = lua_dir,
-      LUA_BINDIR = lua_bindir,
-      LUA_INCDIR = lua_incdir,
-      LUA_LIBDIR = lua_libdir,
-   }
+   -- Use detected values as defaults, overridable via config files or CLI args
+
+   local first_arg = get_first_arg()
+
+   cfg.lua_version = detected.lua_version or hardcoded.LUA_VERSION or _VERSION:sub(5)
+   cfg.lua_interpreter = detected.lua_interpreter or hardcoded.LUA_INTERPRETER or (first_arg and first_arg:gsub(".*[\\/]", "")) or (is_windows and "lua.exe" or "lua")
+   cfg.project_dir = (not hardcoded.FORCE_CONFIG) and detected.project_dir
+
+   do
+      local lua_bindir = detected.lua_bindir or hardcoded.LUA_BINDIR or (first_arg and first_arg:gsub("[\\/][^\\/]+$", ""))
+      local lua_dir = detected.lua_dir or hardcoded.LUA_DIR or (lua_bindir and lua_bindir:gsub("[\\/]bin$", ""))
+      cfg.variables = {
+         LUA_DIR = lua_dir,
+         LUA_BINDIR = lua_bindir,
+         LUA_LIBDIR = hardcoded.LUA_LIBDIR,
+         LUA_INCDIR = hardcoded.LUA_INCDIR,
+      }
+   end
 
    cfg.init = init
 
@@ -624,12 +658,14 @@ function cfg.init(detected, warning)
          -- running from the Development Command prompt for VS 2017
          system = "windows"
       else
-         local fd = io.open("/bin/sh", "r")
-         if fd then
-            fd:close()
+         local msystem = os.getenv("MSYSTEM")
+         if msystem == nil then
+            system = "mingw"
+         elseif msystem == "MSYS" then
             system = "msys"
          else
-            system = "mingw"
+            -- MINGW32 or MINGW64
+            system = "msys2_mingw_w64"
          end
       end
    end
@@ -646,27 +682,37 @@ function cfg.init(detected, warning)
    local sys_config_file
    local home_config_file
    local project_config_file
+
+   local config_file_name = "config-"..cfg.lua_version..".lua"
+
    do
-      set_confdirs(cfg, platforms, hardcoded)
-      local name = "config-"..cfg.lua_version..".lua"
-      sys_config_file = (cfg.sysconfdir .. "/" .. name):gsub("\\", "/")
-      home_config_file = (cfg.homeconfdir .. "/" .. name):gsub("\\", "/")
-      if project_dir then
-         project_config_file = project_dir .. "/.luarocks/" .. name
+      local sysconfdir = os.getenv("LUAROCKS_SYSCONFDIR") or hardcoded.SYSCONFDIR
+      if platforms.windows and not platforms.msys2_mingw_w64 then
+         cfg.home = os.getenv("APPDATA") or "c:"
+         cfg.home_tree = cfg.home.."/luarocks"
+         cfg.sysconfdir = sysconfdir or ((os.getenv("PROGRAMFILES") or "c:") .. "/luarocks")
+      else
+         cfg.home = hardcoded.HOMEDIR or os.getenv("HOME") or ""
+         local localdir = hardcoded.LOCALDIR or cfg.home
+         local home_tree_subdir = hardcoded.HOME_TREE_SUBDIR or "/.luarocks"
+         cfg.homeconfdir = localdir .. home_tree_subdir
+         cfg.home_tree = localdir .. home_tree_subdir
+         cfg.sysconfdir = sysconfdir or detect_sysconfdir() or "/etc/luarocks"
       end
    end
 
    -- Load system configuration file
+   sys_config_file = (cfg.sysconfdir .. "/" .. config_file_name):gsub("\\", "/")
    local sys_config_ok, err = load_config_file(cfg, platforms, sys_config_file)
    if err then
-      return nil, err, "config"
+      exit_ok, exit_err, exit_what = nil, err, "config"
    end
 
    -- Load user configuration file (if allowed)
    local home_config_ok
    local project_config_ok
    if not hardcoded.FORCE_CONFIG then
-      local env_var   = "LUAROCKS_CONFIG_" .. lua_version:gsub("%.", "_")
+      local env_var   = "LUAROCKS_CONFIG_" .. cfg.lua_version:gsub("%.", "_")
       local env_value = os.getenv(env_var)
       if not env_value then
          env_var   = "LUAROCKS_CONFIG"
@@ -676,7 +722,7 @@ function cfg.init(detected, warning)
       if env_value then
          local env_ok, err = load_config_file(cfg, platforms, env_value)
          if err then
-            return nil, err, "config"
+            exit_ok, exit_err, exit_what = nil, err, "config"
          elseif warning and not env_ok then
             warning("Warning: could not load configuration file `"..env_value.."` given in environment variable "..env_var.."\n")
          end
@@ -686,19 +732,33 @@ function cfg.init(detected, warning)
          end
       end
 
-      -- try the alternative defaults if there was no environment specified file or it didn't work
-      if not home_config_ok then
+      -- try XDG config home
+      if platforms.unix and not home_config_ok then
+         local xdg_config_home = os.getenv("XDG_CONFIG_HOME") or cfg.home .. "/.config"
+         cfg.homeconfdir = xdg_config_home .. "/luarocks"
+         home_config_file = (cfg.homeconfdir .. "/" .. config_file_name):gsub("\\", "/")
          home_config_ok, err = load_config_file(cfg, platforms, home_config_file)
          if err then
-            return nil, err, "config"
+            exit_ok, exit_err, exit_what = nil, err, "config"
+         end
+      end
+
+      -- try the alternative defaults if there was no environment specified file or it didn't work
+      if not home_config_ok then
+         cfg.homeconfdir = cfg.home_tree
+         home_config_file = (cfg.homeconfdir .. "/" .. config_file_name):gsub("\\", "/")
+         home_config_ok, err = load_config_file(cfg, platforms, home_config_file)
+         if err then
+            exit_ok, exit_err, exit_what = nil, err, "config"
          end
       end
 
       -- finally, use the project-specific config file if any
-      if project_dir then
+      if cfg.project_dir then
+         project_config_file = cfg.project_dir .. "/.luarocks/" .. config_file_name
          project_config_ok, err = load_config_file(cfg, platforms, project_config_file)
          if err then
-            return nil, err, "config"
+            exit_ok, exit_err, exit_what = nil, err, "config"
          end
       end
    end
@@ -708,15 +768,16 @@ function cfg.init(detected, warning)
    -- Let's finish up the cfg table.
    ----------------------------------------
 
-   -- Settings detected or given via the CLI (i.e. --lua-dir) take precedence over config files:
-   cfg.project_dir = detected.project_dir
-   cfg.lua_version = detected.lua_version or cfg.lua_version
-   cfg.luajit_version = detected.luajit_version or cfg.luajit_version
-   cfg.lua_interpreter = detected.lua_interpreter or cfg.lua_interpreter
-   cfg.variables.LUA_BINDIR = detected.lua_bindir or cfg.variables.LUA_BINDIR or lua_bindir
-   cfg.variables.LUA_INCDIR = detected.lua_incdir or cfg.variables.LUA_INCDIR or lua_incdir
-   cfg.variables.LUA_LIBDIR = detected.lua_libdir or cfg.variables.LUA_LIBDIR or lua_libdir
-   cfg.variables.LUA_DIR = detected.lua_dir or cfg.variables.LUA_DIR or lua_dir
+   -- Settings given via the CLI (i.e. --lua-dir) take precedence over config files.
+   cfg.project_dir = detected.given_project_dir or cfg.project_dir
+   cfg.lua_version = detected.given_lua_version or cfg.lua_version
+   if detected.given_lua_dir then
+      cfg.variables.LUA_DIR = detected.given_lua_dir
+      cfg.variables.LUA_BINDIR = detected.lua_bindir
+      cfg.variables.LUA_LIBDIR = nil
+      cfg.variables.LUA_INCDIR = nil
+      cfg.lua_interpreter = detected.lua_interpreter
+   end
 
    -- Build a default list of rocks trees if not given
    if cfg.rocks_trees == nil then
@@ -729,10 +790,10 @@ function cfg.init(detected, warning)
       end
    end
 
-   local defaults = make_defaults(lua_version, processor, platforms, cfg.home, hardcoded)
+   local defaults = make_defaults(cfg.lua_version, processor, platforms, cfg.home, hardcoded)
 
    if platforms.windows and hardcoded.WIN_TOOLS then
-      local tools = { "SEVENZ", "CP", "FIND", "LS", "MD5SUM", "PWD", "RMDIR", "WGET", "MKDIR" }
+      local tools = { "SEVENZ", "CP", "FIND", "LS", "MD5SUM", "WGET", }
       for _, tool in ipairs(tools) do
          defaults.variables[tool] = '"' .. hardcoded.WIN_TOOLS .. "/" .. defaults.variables[tool] .. '.exe"'
       end
@@ -740,14 +801,13 @@ function cfg.init(detected, warning)
       defaults.fs_use_modules = true
    end
 
-   defaults.rocks_provided, defaults.rocks_provided_3_0 = make_rocks_provided(lua_version, luajit_version)
    use_defaults(cfg, defaults)
 
    cfg.variables.LUA = cfg.variables.LUA or (cfg.variables.LUA_BINDIR and (cfg.variables.LUA_BINDIR .. "/" .. cfg.lua_interpreter):gsub("//", "/"))
    cfg.user_agent = "LuaRocks/"..cfg.program_version.." "..cfg.arch
 
    cfg.config_files = {
-      project = project_dir and {
+      project = cfg.project_dir and {
          file = project_config_file,
          found = not not project_config_ok,
       },
@@ -766,6 +826,8 @@ function cfg.init(detected, warning)
                     or sys_config_file),
    }
 
+   cfg.cache = {}
+
    ----------------------------------------
    -- Attributes of cfg are set.
    -- Let's add some methods.
@@ -807,8 +869,8 @@ function cfg.init(detected, warning)
 
    function cfg.init_package_paths()
       local lr_path, lr_cpath, lr_bin = cfg.package_paths()
-      package.path = util.cleanup_path(package.path .. ";" .. lr_path, ";", lua_version)
-      package.cpath = util.cleanup_path(package.cpath .. ";" .. lr_cpath, ";", lua_version)
+      package.path = util.cleanup_path(package.path .. ";" .. lr_path, ";", cfg.lua_version, true)
+      package.cpath = util.cleanup_path(package.cpath .. ";" .. lr_cpath, ";", cfg.lua_version, true)
    end
 
    --- Check if platform was detected
@@ -819,12 +881,21 @@ function cfg.init(detected, warning)
       return platforms[name]
    end
 
-   function cfg.each_platform()
-      local i = 0
+   -- @param direction (optional) "least-specific-first" (default) or "most-specific-first"
+   function cfg.each_platform(direction)
+      direction = direction or "least-specific-first"
+      local i, delta
+      if direction == "least-specific-first" then
+         i = 0
+         delta = 1
+      else
+         i = #platform_order + 1
+         delta = -1
+      end
       return function()
          local p
          repeat
-            i = i + 1
+            i = i + delta
             p = platform_order[i]
          until (not p) or platforms[p]
          return p
@@ -841,7 +912,7 @@ function cfg.init(detected, warning)
    end
 
    cfg.initialized = true
-   return true
+   return exit_ok, exit_err, exit_what
 end
 
 return cfg
