@@ -10,6 +10,8 @@ local cfg = require("luarocks.core.cfg")
 local dir = require("luarocks.dir")
 local util = require("luarocks.util")
 
+local pack = table.pack or function(...) return { n = select("#", ...), ... } end
+
 local socket_ok, zip_ok, lfs_ok, md5_ok, digest_ok, posix_ok, bz2_ok, _
 local http, ftp, zip, lfs, md5, digest, posix, bz2
 
@@ -55,7 +57,9 @@ end
 
 local function quote_args(command, ...)
    local out = { command }
-   for _, arg in ipairs({...}) do
+   local args = pack(...)
+   for i=1, args.n do
+      local arg = args[i]
       assert(type(arg) == "string")
       out[#out+1] = fs.Q(arg)
    end
@@ -111,25 +115,35 @@ end
 -- The tool is executed using a flag, usually just to ask its version.
 -- @param tool_cmd string: The command to be used to check the tool's presence (e.g. hg in case of Mercurial)
 -- @param tool_name string: The actual name of the tool (e.g. Mercurial)
--- @param arg string: The flag to pass to the tool. '--version' by default.
-function fs_lua.is_tool_available(tool_cmd, tool_name, arg)
+function fs_lua.is_tool_available(tool_cmd, tool_name)
    assert(type(tool_cmd) == "string")
    assert(type(tool_name) == "string")
 
-   arg = arg or "--version"
-   assert(type(arg) == "string")
-
    local ok
    if tool_available_cache[tool_name] ~= nil then
       ok = tool_available_cache[tool_name]
    else
-      ok = fs.execute_quiet(tool_cmd, arg)
+      local tool_cmd_no_args = tool_cmd:gsub(" [^\"]*$", "")
+
+      -- if it looks like the tool has a pathname, try that first
+      if tool_cmd_no_args:match("[/\\]") then
+         local fd = io.open(tool_cmd_no_args, "r")
+         if fd then
+            fd:close()
+            ok = true
+         end
+      end
+
+      if not ok then
+         ok = fs.search_in_path(tool_cmd_no_args)
+      end
+
       tool_available_cache[tool_name] = (ok == true)
    end
 
    if ok then
       return true
-   else   
+   else
       local msg = "'%s' program not found. Make sure %s is installed and is available in your PATH " ..
                   "(or you may want to edit the 'variables.%s' value in file '%s')"
       return nil, msg:format(tool_cmd, tool_name, tool_name:upper(), cfg.config_files.nearest)
@@ -212,7 +226,7 @@ function fs_lua.modules(at)
 
    local modules = {}
    local is_duplicate = {}
-   for _, path in ipairs(paths) do
+   for _, path in ipairs(paths) do  -- luacheck: ignore 421
       local files = fs.list_dir(path)
       for _, filename in ipairs(files or {}) do
          if filename:match("%.[lL][uU][aA]$") then
@@ -381,6 +395,20 @@ function fs_lua.remove_dir_tree_if_empty(d)
    end
 end
 
+local function are_the_same_file(f1, f2)
+   if f1 == f2 then
+      return true
+   end
+   if cfg.is_platform("unix") then
+      local i1 = lfs.attributes(f1, "ino")
+      local i2 = lfs.attributes(f2, "ino")
+      if i1 ~= nil and i1 == i2 then
+         return true
+      end
+   end
+   return false
+end
+
 --- Copy a file.
 -- @param src string: Pathname of source
 -- @param dest string: Pathname of destination
@@ -396,7 +424,7 @@ function fs_lua.copy(src, dest, perms)
    if destmode == "directory" then
       dest = dir.path(dest, dir.base_name(src))
    end
-   if src == dest or (cfg.is_platform("unix") and lfs.attributes(src, "ino") == lfs.attributes(dest, "ino")) then
+   if are_the_same_file(src, dest) then
       return nil, "The source and destination are the same files"
    end
    local src_h, err = io.open(src, "rb")
@@ -518,7 +546,7 @@ end
 --- Internal implementation function for fs.dir.
 -- Yields a filename on each iteration.
 -- @param at string: directory to list
--- @return nil or (nil and string): an error message on failure 
+-- @return nil or (nil and string): an error message on failure
 function fs_lua.dir_iterator(at)
    local pok, iter, arg = pcall(lfs.dir, at)
    if not pok then
@@ -688,9 +716,9 @@ local redirect_protocols = {
    https = luasec_ok and https,
 }
 
-local function request(url, method, http, loop_control)
+local function request(url, method, http, loop_control)  -- luacheck: ignore 431
    local result = {}
-   
+
    if cfg.verbose then
       print(method, url)
    end
@@ -788,13 +816,13 @@ end
 -- via the HTTP Last-Modified header if the full download is needed.
 -- @return (boolean | (nil, string, string?)): True if successful, or
 -- nil, error message and optionally HTTPS error in case of errors.
-local function http_request(url, filename, http, cache)
+local function http_request(url, filename, http, cache)  -- luacheck: ignore 431
    if cache then
       local status = read_timestamp(filename..".status")
       local timestamp = read_timestamp(filename..".timestamp")
       if status or timestamp then
          local unixtime = read_timestamp(filename..".unixtime")
-         if unixtime then
+         if tonumber(unixtime) then
             local diff = os.time() - tonumber(unixtime)
             if status then
                if diff < cfg.cache_fail_timeout then
@@ -807,7 +835,7 @@ local function http_request(url, filename, http, cache)
             end
          end
 
-         local result, status, headers, err = request(url, "HEAD", http)
+         local result, status, headers, err = request(url, "HEAD", http)  -- luacheck: ignore 421
          if not result then
             return fail_with_status(filename, status, headers)
          end
@@ -892,8 +920,11 @@ function fs_lua.download(url, filename, cache)
       err = "Unsupported protocol"
    end
    if https_err then
+      local downloader, err = fs.which_tool("downloader")
+      if not downloader then
+         return nil, err
+      end
       if not downloader_warning then
-         local downloader = fs.which_tool("downloader")
          util.warning("falling back to "..downloader.." - install luasec to get native HTTPS support")
          downloader_warning = true
       end
@@ -1016,6 +1047,10 @@ function fs_lua.current_user()
    return posix.getpwuid(posix.geteuid()).pw_name
 end
 
+function fs_lua.is_superuser()
+   return posix.geteuid() == 0
+end
+
 -- This call is not available on all systems, see #677
 if posix.mkdtemp then
 
@@ -1095,10 +1130,10 @@ end
 
 --- Check if user has write permissions for the command.
 -- Assumes the configuration variables under cfg have been previously set up.
--- @param flags table: the flags table passed to run() drivers.
+-- @param args table: the args table passed to run() drivers.
 -- @return boolean or (boolean, string): true on success, false on failure,
 -- plus an error message.
-function fs_lua.check_command_permissions(flags)
+function fs_lua.check_command_permissions(args)
    local ok = true
    local err = ""
    for _, directory in ipairs { cfg.rocks_dir, cfg.deploy_lua_dir, cfg.deploy_bin_dir, cfg.deploy_lua_dir } do
@@ -1127,7 +1162,7 @@ function fs_lua.check_command_permissions(flags)
    if ok then
       return true
    else
-      if flags["local"] or cfg.local_by_default then
+      if args["local"] or cfg.local_by_default then
          err = err .. " \n-- please check your permissions."
       else
          err = err .. " \n-- you may want to run as a privileged user or use your local tree with --local."
